/***********************************************************
 *
 * This script (3 of 3) is part of the following presentation:
 *   Unique, Pseudo Random Numbers: Generate without using RAM, Disk, CPU, I/O, or even time
 *
 * Script:     UniquePseudoRandomNumbers_03_MultiplicativeInverseTests.sql
 * Created by: Solomon Rutzky  ( http://www.SqlQuantumLeap.com/ )
 * Company:    Sql Quantum Lift  ( http://www.SqlQuantumLift.com/ )
 * Created on: 2016-07-10
 * Updated on: 2017-06-19
 *
 * The latest version of this script can be found at:
 * http://pastebin.com/7u2NzYtd
 *
 ***********************************************************/
IF (OBJECT_ID(N'tempdb..#NoF5') IS NULL)
BEGIN
    CREATE TABLE #NoF5 (Col1 INT);
END;
GO
SET PARSEONLY ON;
GO
---------------------------------------------------------------------------------

--
----- This script is NOT intended to be run all at once (i.e. F5).
----- You should highlight each statement, or at least each section
----- (from BEGIN to END), to run individually.
--
----- When connected to SQL Server, each section can be collapsed
----- (i.e. hidden) by clicking the "-" to the left of each "BEGIN".
--
GO

-- =================================================
-- SETUP (run once)
USE [tempdb];
SET NOCOUNT, ANSI_NULLS, ANSI_PADDING, QUOTED_IDENTIFIER ON;
GO

-- =================================================
BEGIN -- =  Greatest Common Divisor (GCD)          =
PRINT 'Greatest Common Divisor';

-- It is not required that you use, or even create, the
-- GreatestCommonDivisor function. It is used here to help
-- test the basic mathematical concept since it is required
-- that the result be 1 in order for the numbers to work
-- with the Modular Multiplicative Inverse alogrithm.

SELECT MMI.GreatestCommonDivisor(5, 4); -- 1

SELECT MMI.GreatestCommonDivisor(5, 10); -- 5

SELECT MMI.GreatestCommonDivisor(10, 5); -- 5

SELECT MMI.GreatestCommonDivisor(42, 2017); -- 1

----------------------------------------------------
END
GO



-- =================================================
BEGIN -- =  Modular Multiplicative Inverse         =
PRINT 'Modular Multiplicative Inverse';

-- MMI.ModularMultiplicativeInverse(_any_value_coprime_to_upper_limit_, _upper_limit_)
-- coprime means: GCD(x, y) = 1

SELECT MMI.GreatestCommonDivisor(42, 2017); -- 1
SELECT MMI.ModularMultiplicativeInverse(42, 2017); -- 1969
SELECT (42 * 1969) % 2017; -- 1


SELECT MMI.ModularMultiplicativeInverse(40, 1); -- 0

SELECT MMI.ModularMultiplicativeInverse(52, -217); -- 96

SELECT MMI.ModularMultiplicativeInverse(-486, 217); -- 121

SELECT MMI.ModularMultiplicativeInverse(40, 2018); -- -1

SELECT MMI.ModularMultiplicativeInverse(2700, 53); -- 35

----------------------------------------------------
END
GO



-- =================================================
BEGIN -- =  Not so much working                    =

SELECT MMI.GreatestCommonDivisor(2, 10); -- 2
SELECT MMI.ModularMultiplicativeInverse(2, 10); -- -1

----------------------------------------------------
END
GO



-- =================================================
BEGIN -- =  Simple Verification                    =
PRINT 'Simple Verification';

-- example values from https://ericlippert.com/2013/11/14/a-practical-use-of-multiplicative-inverses/
SELECT MMI.ModularMultiplicativeInverse(387420489, 1000000000); -- 513180409

SELECT MMI.Encode(387420489, 1000000000, 35); -- 559717115

SELECT MMI.Decode(513180409, 1000000000, 559717115); -- 35

----------------------------------------------------
END
GO



/*
	For the following tests, pay close attention to the "Encoded"
	field -- output of the MMI.Decode function -- as it shows the
	non-sequential, pseudo-random result of this method.

	The "Decoded" field -- output of the MMI.Decode function -- shows
	that you can always get back to the original value of any "encoded"
	number, so long as you use the same base and MMI values.

	The "DifferenceFromPrior" field shows the row-by-row change so
	that it is easier to see that there is a pattern and what it is.
*/
GO

-- =================================================
BEGIN -- =  Small set to see pattern               =

SELECT MMI.GreatestCommonDivisor(7, 11); -- 1
SELECT MMI.ModularMultiplicativeInverse(7, 11); -- 8

SELECT * FROM MMI.GenerateTest(7, 11, 8, 10);

----------------------------------------------------
END
GO



-- =================================================
BEGIN -- =  Set of 20 to help see behavior         =

SELECT MMI.GreatestCommonDivisor(5, 21); -- 1
SELECT MMI.ModularMultiplicativeInverse(5, 21); -- 17

SELECT * FROM MMI.GenerateTest(5, 21, 17, 20);

----------------------------------------------------
END
GO



-- =================================================
BEGIN -- =  Set of 20 with different MMI           =

SELECT MMI.GreatestCommonDivisor(5, 211); -- 1
SELECT MMI.ModularMultiplicativeInverse(5, 211); -- 169

SELECT * FROM MMI.GenerateTest(5, 211, 169, 20);

----------------------------------------------------
END
GO



-- =================================================
BEGIN -- =  Set of 20 with different base and MMI  =

SELECT MMI.GreatestCommonDivisor(131, 211); -- 1
SELECT MMI.ModularMultiplicativeInverse(131, 211); -- 29

SELECT * FROM MMI.GenerateTest(131, 211, 29, 20);

----------------------------------------------------
END
GO



-- =================================================
BEGIN -- =  Compare different MMI's for same base  =

SELECT MMI.GreatestCommonDivisor(5, 21); -- 1
SELECT MMI.ModularMultiplicativeInverse(2, 21); -- 11
SELECT MMI.ModularMultiplicativeInverse(4, 21); -- 15
SELECT MMI.ModularMultiplicativeInverse(5, 21); -- 17

;WITH nums(Num) AS
(
	SELECT ROW_NUMBER() OVER (ORDER BY (SELECT 0))
	FROM   [master].[sys].[objects]
)
SELECT	TOP (20) nm.[Num],
		MMI.Encode(2, 21, nm.[Num]) AS [Encoded-2],
		MMI.Encode(4, 21, nm.[Num]) AS [Encoded-4],
		MMI.Encode(5, 21, nm.[Num]) AS [Encoded-5]
FROM		nums nm
ORDER BY	nm.[Num];

----------------------------------------------------
END
GO



-- =================================================
BEGIN -- =  Set of 60,000 with 2 combinations      =

SELECT MMI.ModularMultiplicativeInverse(59, 86400); -- 39539
SELECT * FROM MMI.GenerateTest(59, 86400, 39539, 60000);

SELECT MMI.ModularMultiplicativeInverse(5977, 86400); -- 30313
SELECT * FROM MMI.GenerateTest(5977, 86400, 30313, 60000);

----------------------------------------------------
END
GO



-- =================================================
BEGIN -- =  Set of 60,000 with big upper-limit     =

SELECT MMI.ModularMultiplicativeInverse(13107, 10000000); -- 1045243
SELECT * FROM MMI.GenerateTest(13107, 10000000, 1045243, 60000);

SELECT MMI.ModularMultiplicativeInverse(131079, 10000000); -- 2430519
SELECT * FROM MMI.GenerateTest(131079, 10000000, 2430519, 60000);

----------------------------------------------------
END
GO



-- =================================================
BEGIN -- =  Same 60k and upper-limit, but better   =
PRINT 'Same 60k and upper-limit, but better';

-- This is the same 60,000 item set as the test above,
-- and the same upper-limit that is much larger than
-- the set itself, but the per-row variation is much
-- better and looks much more random. So keep trying
-- different combinations until something works.

SELECT MMI.ModularMultiplicativeInverse(17000903, 10000000); -- 4696567
SELECT * FROM MMI.GenerateTest(17000903, 10000000, 4696567, 60000);

----------------------------------------------------
END
GO


BEGIN

SET PARSEONLY OFF;
IF (OBJECT_ID(N'tempdb..#NoF5') IS NOT NULL)
BEGIN
    DROP TABLE #NoF5;
	RAISERROR(N'Please do not execute this entire script (i.e. F5, Control-E, etc.)', 16, 1);
	RETURN;
END;

END
